<?php

namespace Plugins\Revisions;

require_once __DIR__ . '/vendor/autoload.php';

use Psr\Http\Message\ServerRequestInterface as Request;
use Psr\Http\Message\ResponseInterface as Response;
use Typemill\Plugin;
use Typemill\Models\Content;
use Typemill\Models\Navigation;
use Jfcherng\Diff\DiffHelper;
use Jfcherng\Diff\Factory\RendererFactory;

class Revisions extends Plugin
{
    # more feature ideas: 
    # delete all revisions
    # add versions to revisions or create versions manually (simply add version number in plugin)
    # add options show line, show word, show character
    # add option length of text or all.

    private $settings;

    private $storagePath;

    public static function setPremiumLicence()
    {
        return 'BUSINESS';
    }   

    public static function getSubscribedEvents()
    {
        return array(
            'onTwigLoaded'          => 'onTwigLoaded',
            'onPageUpdated'         => 'onPageUpdated', # done
            'onPagePublished'       => 'onPagePublished', # create a new revision
            'onPageDiscard'         => 'onPageUpdated', # like updated, only relevant for change mode 
            'onPageRenamed'         => 'onPageRenamed', # change url
            'onPageSorted'          => 'onPageSorted', # change url
            'onPageDeleted'         => 'onPageDeleted', # delete revisions or option to store revisions for deleted pages
        );
    }

    public static function addNewRoutes()
    {
        return [
            [
                'httpMethod'    => 'get', 
                'route'         => '/api/v1/revisions', 
                'name'          => 'revisions.get', 
                'class'         => 'Plugins\revisions\revisions:getRevisions', 
                'resource'      => 'mycontent', 
                'privilege'     => 'write'
            ],
            [
                'httpMethod'    => 'put', 
                'route'         => '/api/v1/revision',
                'name'          => 'revision.restore', 
                'class'         => 'Plugins\revisions\revisions:restoreRevision', 
                'resource'      => 'mycontent', 
                'privilege'     => 'write'
            ],            
            [
                'httpMethod'    => 'get', 
                'route'         => '/api/v1/revision', 
                'name'          => 'revision.get', 
                'class'         => 'Plugins\revisions\revisions:getRevision', 
                'resource'      => 'mycontent', 
                'privilege'     => 'write'
            ]
        ];
    }

    public function onTwigLoaded($noidea)
    {
        if($this->editorroute)
        {
            $this->addJS('/revisions/js/revisions.js');
            $this->addCSS('/revisions/css/revisions.css');
        }
    }

    public function onPageUpdated($data)
    {
        $this->settings     = $this->getSettings();
        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        # only apply for mode 'changes';
        if(isset($this->settings['plugins']['revisions']['mode']) && $this->settings['plugins']['revisions']['mode'] == 'versions')
        {
            return false;
        }

        $pageData       = $data->getData();
        $item           = $pageData['item'];
        $oldMarkdown    = $pageData['oldMarkdown'];
        $newMarkdown    = $pageData['newMarkdown'];
        $username       = $pageData['username'];

        $filename       = $this->getFileName($item);
        $filepath       = $this->storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        $revisions      = $this->getPageRevisions($filepath);

        if(!$revisions)
        {
            # avoid that the initial revision has the same timestamp, because revisions are searched with timesstamp
            $timestamp = time() - 500;

            # create an initial revision with the old content
            $revisions = $this->addRevision([], $oldMarkdown, 'initial', 'draft', $timestamp);

            # add another revision with the user and the new content
            $revisions = $this->addRevision($revisions, $newMarkdown, $username, 'draft');

            $this->storeRevisions($revisions, $filepath);
        }
        else
        {
            $lastRevision       = end($revisions);
            $lastUsername       = $lastRevision['username'];
            $lastTimestamp      = $lastRevision['timestamp'];
            $status             = $lastRevision['status'];

            $configuredPeriod   = $this->settings['plugins']['revisions']['peruser'] ?? 1;
            $configuredPeriod   = $configuredPeriod * 86400;

            # if it is the same user and the same day
            if($status != 'live' && $username == $lastUsername && (time() - $lastTimestamp) < $configuredPeriod)
            {
                # replace the last revision
                array_pop($revisions);
            }

            # add the new revision (in case same user and same day it will replace the last reviision)
            $revisions = $this->addRevision($revisions, $newMarkdown, $username, 'draft');

            $this->storeRevisions($revisions, $filepath);
        }
    }

    public function onPagePublished($data)
    {
        $this->settings     = $this->getSettings();
        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        $pageData       = $data->getData();
        $item           = $pageData['item'] ?? false; 
        $markdown       = $pageData['markdown'] ?? false;
        $username       = $pageData['username'] ?? false;

        $filename       = $this->getFileName($item);
        $filepath       = $this->storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        $revisions      = $this->getPageRevisions($filepath);

        if(!$revisions)
        {
            # add another revision with the user and the new content
            $revisions = $this->addRevision($revisions, $markdown, $username, 'live');

            $this->storeRevisions($revisions, $filepath);
        }
        else
        {
            $lastRevision       = end($revisions);
            $lastUsername       = $lastRevision['username'];
            $lastTimestamp      = $lastRevision['timestamp'];

            $configuredPeriod   = $this->settings['plugins']['revisions']['peruser'] ?? 1;
            $configuredPeriod   = $configuredPeriod * 86400;

            # if it is the same user and the same day
            if($username == $lastUsername && (time() - $lastTimestamp) < $configuredPeriod)
            {
                # replace the last revision
                array_pop($revisions);
            }

            # add the new revision (in case same user and same day it will replace the last reviision)
            $revisions = $this->addRevision($revisions, $markdown, $username, 'live');

            $this->storeRevisions($revisions, $filepath);
        }
    }

    public function onPageDeleted($data)
    {
        $this->settings     = $this->getSettings();

        if(isset($this->settings['plugins']['revisions']['keepRevisions']) && $this->settings['plugins']['revisions']['keepRevisions'] == true)
        {
            # optiion not delete old revisions but could lead to confusions if page with same name again (shows old revisions)
            return false;
        }

        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        $item = $data->getData();

        $filename       = $this->getFileName($item);
        $filepath       = $this->storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        if(file_exists($filepath))
        {
            unlink($filepath);
        }
    }

    public function onPageRenamed($data)
    {
        $this->settings     = $this->getSettings();
        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        $pagedata   = $data->getData();
 
        $olditem       = $pagedata['item'] ?? false;
        $newurl        = $pagedata['newUrl'] ?? false;

        if(!$olditem && !$newurl)
        {
            return false;
        }

        $oldfilename    = $this->getFileName($olditem);
        $newfilename    = $this->getFilenameFromUrl($newurl);

        $oldfiles       = glob($this->storagePath . DIRECTORY_SEPARATOR . $oldfilename . '*');

        foreach($oldfiles as $oldfile)
        {
            $newfile = str_replace($oldfilename, $newfilename, $oldfile);
            
            if (file_exists($oldfile))
            {
                rename($oldfile, $newfile);
            }
        }
    }

    public function onPageSorted($data)
    {
        $pagedata = $data->getData();  // Assume $pagedata is an object

        $olditem  = $pagedata['olditem'] ?? false;
        $newitem  = $pagedata['newitem'] ?? false;

        # if url of the page has changed...
        if(
            $olditem 
            && $newitem
            && isset($olditem->urlRelWoF) 
            && isset($newitem->urlRelWoF)
            && $olditem->urlRelWoF != $newitem->urlRelWoF
        )
        {
            $this->settings     = $this->getSettings();
            $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

            $newfilename        = $this->getFilenameFromUrl($newitem->urlRelWoF);
            $oldfilename        = $this->getFilenameFromUrl($olditem->urlRelWoF);
            $oldfiles           = glob($this->storagePath . DIRECTORY_SEPARATOR . $oldfilename . '*');

            # glob even not needed because you cannot move a folder to another folder
            foreach($oldfiles as $oldfile)
            {
                $newfile = str_replace($oldfilename, $newfilename, $oldfile);
                
                if (!rename($oldfile, $newfile))
                {
                    echo "Failed to rename $oldfile to $newfile";
                }
            }
        }
    }

    public function getRevisions(Request $request, Response $response, $args)
    {
        $params             = $request->getQueryParams();
        $itemUrlWoF         = $params['itemurl'] ?? false;

        if(!$itemUrlWoF OR $itemUrlWoF == '')
        {
            $response->getBody()->write(json_encode([
                'message'   => 'Itemurl is missing.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);
        }

        $this->settings     = $this->getSettings();
        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        $filename           = $this->getFilenameFromUrl($itemUrlWoF);
        $filepath           = $this->storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        $revisions          = $this->getPageRevisions($filepath);
        if($revisions && is_array($revisions))
        {
            $maxVersions        = $settings['plugins']['revisions']['perpage'] ?? 10;
            if(count($revisions) > $maxVersions)
            {
                $revisions = array_slice($revisions, -$maxVersions);

                $this->storeRevisions($revisions, $filepath);
            }

            $revisions          = array_reverse($revisions);
            foreach($revisions as $key => $revision)
            {
                $revisions[$key]['key']     = $revision['username'] . $key;
                $revisions[$key]['date']    = date('d.m.Y', $revision['timestamp']);                
            }
        }

        $response->getBody()->write(json_encode([
            'revisions' => $revisions
        ]));

        return $response->withHeader('Content-Type', 'application/json');       
    }

    public function getRevision(Request $request, Response $response, $args)
    {
        $params             = $request->getQueryParams();
        $itemUrlWoF         = $params['itemurl'] ?? false;
        $timestamp          = $params['timestamp'] ?? false;

        if(!$itemUrlWoF || $itemUrlWoF == '' ||  !$timestamp)
        {
            $response->getBody()->write(json_encode([
                'message'   => 'Itemurl or timestamp are missing.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);
        }

        $this->settings     = $this->getSettings();
        $this->storagePath  = $this->settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';

        $filename           = $this->getFilenameFromUrl($itemUrlWoF);
        $filepath           = $this->storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        $revisions          = $this->getPageRevisions($filepath);

        $foundRevision      = false;
        $foundKey           = false;

        foreach($revisions as $key => $revision)
        {
            if($revision['timestamp'] == $timestamp)
            {
                $foundRevision = $revision;
                $foundKey = $key;
            }
        }

        if(!$foundRevision)
        {
            $response->getBody()->write(json_encode([
                'message'   => 'We did not find a revisions for the timestamp.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);
        }

        $olderVersionKey = $foundKey - 1;
        if(isset($revisions[$olderVersionKey]))
        {
            $olderVersion = $revisions[$olderVersionKey];
        }
        else
        {
            $olderVersion = $foundRevision;
        }

        # renderer class name:
        # Text renderers: Context, JsonText, Unified
        # HTML renderers: Combined, Inline, JsonHtml, SideBySide
        $rendererName = 'Combined';
        $differOptions = [
            # show how many neighbor lines
            # Differ::CONTEXT_ALL can be used to show the whole file
            'context' => 200,
            # ignore case difference
            'ignoreCase' => false,
            # ignore line ending difference
            'ignoreLineEnding' => false,
            # ignore whitespace difference
            'ignoreWhitespace' => false,
            # if the input sequence is too long, it will just gives up (especially for char-level diff)
            'lengthLimit' => 2000,
            # if truthy, when inputs are identical, the whole inputs will be rendered in the output
            'fullContextIfIdentical' => true,
        ];
        $rendererOptions = [
            'detailLevel'   => 'word',
            'showHeader'    => false,
        ];

        $diff = DiffHelper::calculate(
            $olderVersion['content'], 
            $foundRevision['content'],
            $rendererName, 
            $differOptions, 
            $rendererOptions
        );
        
        $response->getBody()->write(json_encode([
            'diff' => $diff
        ]));

        return $response->withHeader('Content-Type', 'application/json');       
    }

    public function restoreRevision(Request $request, Response $response, $args)
    {
        $params             = $request->getParsedBody();
        $item               = $params['item'] ?? false;
        $itemUrlWoF         = isset($item['urlRelWoF']) ? $item['urlRelWoF'] : false;
        $timestamp          = $params['timestamp'] ?? false;

        if(!$item || !isset($item['urlRelWoF']) || !$timestamp)
        {
            $response->getBody()->write(json_encode([
                'message'   => 'Item or timestamp are missing.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);
        }
        $item               = (object) $item;
        $settings           = $this->getSettings();
        $storagePath        = $settings['rootPath'] . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'revisions';
        $username           = $request->getAttribute('c_username');

        $filename           = $this->getFilenameFromUrl($itemUrlWoF);
        $filepath           = $storagePath . DIRECTORY_SEPARATOR . $filename . '.json';

        $revisions          = $this->getPageRevisions($filepath);

        $foundRevision      = false;
        $foundKey           = false;

        foreach($revisions as $key => $revision)
        {
            if($revision['timestamp'] == $timestamp)
            {
                $foundRevision = $revision;
                $foundKey = $key;

                # remove the old revision
#                unset($revisions[$key]);

                # add as new revision
                $revisions = $this->addRevision($revisions, $foundRevision['content'], $username, 'draft');
            }
        }

        if(!$foundRevision)
        {
            $response->getBody()->write(json_encode([
                'message'   => 'We did not find a revisions for the timestamp.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);
        }

        # save draft content
        $content            = new Content($this->urlinfo['baseurl'], $settings, $this->container->get('dispatcher'));
        $oldMarkdown        = $content->getDraftMarkdown($item);
        $markdownArray      = $content->markdownTextToArray($foundRevision['content']);
        $result = $content->saveDraftMarkdown($item, $markdownArray);

        if(!$result)
        {
            $response->getBody()->write(json_encode([
                'message'   => 'We could not restore the revision.'
            ]));

            return $response->withHeader('Content-Type', 'application/json')->withStatus(422);            
        }

        $maxVersions        = $settings['plugins']['revisions']['perpage'] ?? 10;
        if(count($revisions) > $maxVersions)
        {
            $revisions = array_slice($revisions, -$maxVersions);
        }

        $this->storeRevisions($revisions, $filepath);

        $revisions          = array_reverse($revisions);
        foreach($revisions as $key => $revision)
        {
            $revisions[$key]['key']     = $revision['username'] . $key;
            $revisions[$key]['date']    = date('d.m.Y', $revision['timestamp']);                
        }

        # we have to refresh the navigation
        $navigation         = new Navigation();
        $naviFileName       = $navigation->getNaviFileNameForPath($item->path);
        $navigation->clearNavigation([$naviFileName]);
       
        $response->getBody()->write(json_encode([
            'revisions' => $revisions
        ]));

        return $response->withHeader('Content-Type', 'application/json');       
    }

    private function getFileName($item)
    {
        if($item && isset($item->urlRelWoF))
        {
            return $this->getFilenameFromUrl($item->urlRelWoF);
        }

        return false;
    }

    private function getFilenameFromUrl($itemUrlWoF)
    {
        $itemUrlWoF = trim($itemUrlWoF, '/');
        return str_replace('/', '_', $itemUrlWoF);
    }

    private function getPageRevisions($filepath)
    {
        # Ensure the directory exists
        if (!is_dir(dirname($filepath)))
        {
            mkdir(dirname($filepath), 0755, true);
        }

        # Ensure the file exists
        if (!file_exists($filepath))
        {
            touch($filepath);
            chmod($filepath, 0644);
        
            return false;
        }

        $revisions = json_decode(file_get_contents($filepath), $array = true);

        return $revisions;
    }

    private function addRevision(array $revisions, string $markdown, string $username, $status = NULL, $time = NULL)
    {
        # Store the new revision with metadata
        $revision = [
            'timestamp'     => $time ?? time(),
            'status'        => $status,
            'username'      => $username,
            'content'       => $markdown
        ];

        $revisions[] = $revision;

        return $revisions;
    }

    private function storeRevisions(array $revisions, string $filepath)
    {
        file_put_contents($filepath, json_encode($revisions, JSON_PRETTY_PRINT));
    }
}